"""Module to support calculation of metrics of circadian rhythm from acc data"""

import gzip
import numpy as np
import pandas as pd
import pytz
import sys
import scipy as sp
from scipy import fftpack
from datetime import timedelta



def calculatePSD(e, epochPeriod, fourierWithAcc, labels, summary):
    """Calculate the power spectral density from fourier analysis of a 1 day frequency
    
    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param int epochPeriod: Size of epoch time window (in seconds)
    :param bool fourierWithAcc:True calculates fourier done with acceleration data instead of sleep data
    :param list(str) labels: Activity state labels
    :param dict summary: Output dictionary containing all summary metrics

    :return: Write dict <summary> keys 'PSD-<W/Hz>'
    """
    if fourierWithAcc:
        y = e['accImputed'].values
    else:
        cols = [label+'Imputed' for label in labels]  # 'sleepImputed', 'sedentaryImputed', etc...
        y = e[cols].idxmax(axis=1) == 'sleepImputed'  # is sleepImputed highest?
        y = y.values.astype('int')  # e.g. [0,0,0,1,1,0,0,1,...]
        y = 2 * y - 1  # center the signal, [0,1] -> [-1,1]
        
    n = len(y)
    k = len(y)*epochPeriod/(60*60*24)
    a = -2.j * np.pi * k * np.arange(n) / n
    # Find the power spectral density for a one day cycle using fourier analysis 
    res = np.sum(np.exp(a) * y) / n
    PSD = np.abs(res)**2
    summary['PSD'] = PSD



def calculateFourierFreq(e, epochPeriod, fourierWithAcc, labels, summary):
    """Calculate the most prevalent frequency in a fourier analysis 
    
    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param int epochPeriod: Size of epoch time window (in seconds)
    :paran bool fourierWithAcc: True calculates fourier done with acceleration data instead of sleep data
    :param list(str) labels: Activity state labels
    :param dict summary: Output dictionary containing all summary metrics

    :return: Write dict <summary> keys 'fourier frequency-<1/days>'
    """
    if fourierWithAcc:
        y = e['accImputed'].values
    else:
        cols = [label+'Imputed' for label in labels]  # 'sleepImputed', 'sedentaryImputed', etc...
        y = e[cols].idxmax(axis=1) == 'sleepImputed'  # is sleepImputed highest?
        y = y.values.astype('int')  # e.g. [0,0,0,1,1,0,0,1,...]
        y = 2 * y - 1  # center the signal, [0,1] -> [-1,1]
        
    # Fast fourier transform of the sleep column 
    fft_y = np.abs(fftpack.fft(y))
    
    i =  np.arange(1,len(fft_y)) 
    k_max = np.argmax(fft_y[i]) + 1
    n = len(y)
    # Maximise the fourier transform function (func) using the fft_y as a first esitmate 
    func = lambda k: -np.abs(np.sum(np.exp(-2.j * np.pi * k * np.arange(n) / n) * y) / n)
    res = sp.optimize.minimize_scalar(func, bracket=(k_max-1,k_max+1)) 
    # Adjust the frequency to have the units 1/days
    freq_mx = float(res.x)/(len(y)*epochPeriod/(60*60*24))
    summary['fourier-frequency'] = freq_mx


   
def calculateM10L5(e, epochPeriod, summary):
    """Calculates the M10 L5 relative amplitude from the average acceleration from
    the ten most active hours and 5 least most active hours 
    
    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param int epochPeriod: Size of epoch time window (in seconds)
    :param dict summary: Output dictionary containing all summary metrics

    :return: Write dict <summary> keys 'M10 L5-<rel amp>'
    """
    TEN_HOURS = int(10*60*60/epochPeriod)
    FIVE_HOURS = int(5*60*60/epochPeriod)
    num_days = (e.index[-1] - e.index[0]).days
           
    days_split = []
    for n in range(num_days):
        #creates a new list which is used to identify the 24 hour periods in the data frame
        days_split += [n for x in e.index if e.index[0] + timedelta(days=n) <= x < e.index[0] + timedelta(days=n+1)]
    dct = {}
    for i in range(num_days):
        #create new lists with the acceleration data from each 24 hour period
        dct['day_%s' %i] = [e.loc[:,'accImputed'][n] for n in range(len(days_split)) if days_split[n]==i]    
    dct_10 = {}
    dct_5 = {}
    for i in dct:
        #  sums each 10 or 5 hour window with steps of 30s for each day
        dct_10['%s' %i] = [sum(dct['%s' %i][j:j+TEN_HOURS]) for j in range(len(dct['%s' %i])-TEN_HOURS)]
        dct_5['%s' %i] = [sum(dct['%s' %i][j:j+FIVE_HOURS]) for j in range(len(dct['%s' %i])-FIVE_HOURS)]
    avg_10 = {}
    avg_5 = {}
    #   average acceleration (for each 30s) for the max and min windows        
    for i in dct:
        avg_10['%s' %i ] = (np.max(dct_10['%s' %i]))/TEN_HOURS
    for i in dct:
        avg_5['%s' %i] = (np.min(dct_5['%s' %i]))/FIVE_HOURS

    if num_days > 0:
        M10 = sum(avg_10.values())/num_days
        L5 = sum(avg_5.values())/num_days
        rel_amp = (M10-L5)/(M10+L5)
    if num_days < 1:
        rel_amp = 'NA_too_few_days'
    summary['M10L5'] = rel_amp


"""Script to plot accelerometer traces."""

import matplotlib
matplotlib.use('Agg')
from accelerometer import accUtils
import argparse
from datetime import datetime, timedelta, time
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import numpy as np
import os
import pandas as pd
import sys

# http://pandas-docs.github.io/pandas-docs-travis/whatsnew/v0.21.1.html#restore-matplotlib-datetime-converter-registration
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

DOHERTY_NatComms_COLOURS = {'sleep':'blue', 'sedentary':'red',
    'tasks-light':'darkorange', 'walking':'lightgreen', 'moderate':'green'}

WILLETS_SciReports_COLOURS = {'sleep':'blue', 'sit.stand':'red',
    'vehicle':'darkorange', 'walking':'lightgreen', 'mixed':'green',
    'bicycling':'purple'}

WALMSLEY_Nov2020_COLOURS = {'sleep':'blue', 'sedentary':'red',
    'light':'darkorange', 'MVPA':'green'}


def main():
    """
    Application entry point responsible for parsing command line requests
    """

    parser = argparse.ArgumentParser(
        description="A script to plot acc time series data.", add_help=True)
    # required
    parser.add_argument('timeSeriesFile', metavar='input file', type=str,
                            help="input .csv.gz time series file to plot")
    parser.add_argument('--plotFile', metavar='output file', type=str,
                            help="output .png file to plot to")
    parser.add_argument('--activityModel', type=str,
                            default="walmsley",
                            help="""trained activity model .tar file""")
    parser.add_argument('--useRecommendedImputation',
                            metavar='True/False', default=True, type=str2bool,
                            help="""Highly recommended method to show imputed
                            missing data (default : %(default)s)""")
    parser.add_argument('--imputedLabels',
                            metavar='True/False', default=False, type=str2bool,
                            help="""If activity classification during imputed
                            period will be displayed (default : %(default)s)""")
    parser.add_argument('--imputedLabelsHeight',
                            metavar='Proportion i.e. 0-1.0', default=0.9,
                            type=float, help="""Proportion of plot labels take
                            if activity classification during imputed
                            period will be displayed (default : %(default)s)""")
    parser.add_argument('--showFileName',
                            metavar='True/False', default=False, type=str2bool,
                            help="""Toggle showing filename as title in output
                            image (default : %(default)s)""")
    parser.add_argument('--showFirstNDays',
                            metavar='days', default=None,
                            type=int, help="Show just first n days")

    # check input is ok
    if len(sys.argv) < 2:
        msg = "\nInvalid input, please enter at least 1 parameter, e.g."
        msg += "\npython accPlot.py timeSeries.csv.gz \n"
        accUtils.toScreen(msg)
        parser.print_help()
        sys.exit(-1)
    args = parser.parse_args()

    # determine output file name
    if args.plotFile is None:
        inputFileFolder, inputFileName = os.path.split(args.timeSeriesFile)
        inputFileName = inputFileName.split('.')[0]  # remove any extension
        args.plotFile = os.path.join(inputFileFolder, inputFileName + "-plot.png")


    # and then call plot function
    plotTimeSeries(args.timeSeriesFile, args.plotFile,
        showFirstNDays=args.showFirstNDays,
        activityModel=args.activityModel,
        useRecommendedImputation=args.useRecommendedImputation,
        imputedLabels=args.imputedLabels,
        imputedLabelsHeight=args.imputedLabelsHeight,
        showFileName=args.showFileName)



def plotTimeSeries(
        tsFile,
        plotFile,
        showFirstNDays=None,
        activityModel="walmsley",
        useRecommendedImputation=True,
        imputedLabels=False,
        showFileName=False,
        imputedLabelsHeight=0.9):
    """Plot overall activity and classified activity types

    :param str tsFile: Input filename with .csv.gz time series data
    :param str tsFile: Output filename for .png image
    :param int showFirstNDays: Only show first n days of time series (if specified)
    :param str activityModel: Input tar model file used for activity classification
    :param bool useRecommendedImputation: Highly recommended method to show
        imputed values for missing data
    :param bool imputedLabels: If activity classification during imputed period
        will be displayed
    :param float imputedLabelsHeight: Proportion of plot labels take up if
        <imputedLabels> is True
    :param float showFileName: Toggle showing filename as title in output image

    :return: Writes plot to <plotFile>
    :rtype: void

    :Example:
    >>> import accPlot
    >>> accPlot.plotTimeSeries("sample-timeSeries.csv.gz", "sample-plot.png")
    <plot file written to sample-plot.png>
    """

    # read time series file to pandas DataFrame
    d = pd.read_csv(
        tsFile, index_col='time',
        parse_dates=['time'], date_parser=accUtils.date_parser
    )
    if showFirstNDays is not None:
        d = d.first(str(showFirstNDays) + 'D')

    d['acc'] = d['acc'].rolling(window=12, min_periods=1).mean() # smoothing
    d['time'] = d.index.time
    ymin = d['acc'].min()
    ymax = d['acc'].max()

    # infer labels
    labels = []
    for col in d.columns.tolist():
        if col not in [accUtils.TIME_SERIES_COL, 'imputed', 'acc', 'MET']:
            labels += [col]
    print(labels)
    if 'walmsley' in activityModel:
        labels_as_col = WALMSLEY_Nov2020_COLOURS
    elif 'doherty' in activityModel:
        labels_as_col = DOHERTY_NatComms_COLOURS
    elif 'willetts' in activityModel:
        labels_as_col = WILLETS_SciReports_COLOURS
    # add imputation label colour
    labels_as_col['imputed'] = '#fafc6f'

    convert_date = np.vectorize(lambda day, x: matplotlib.dates.date2num(datetime.combine(day, x)))

    # number of rows to display in figure (all days + legend)
    d['date'] = d.index.date
    if not useRecommendedImputation:
        d = d[d['imputed']==0] # if requested, do not show imputed values

    if imputedLabels:
        labelsPosition = imputedLabelsHeight
    else:
        labelsPosition = 1

    groupedDays = d[['acc','time','imputed'] + labels].groupby(by=d['date'])
    nrows = len(groupedDays) + 1

    # create overall figure
    fig = plt.figure(1, figsize=(10,nrows), dpi=100)
    if showFileName:
        plt.suptitle(tsFile)

    # create individual plot for each day
    i = 0
    ax_list = []
    for day, group in groupedDays:
        # set start and end to zero to avoid diagonal fill boxes
        group['imputed'].values[0] = 0
        group['imputed'].values[-1] = 0

        # retrieve time series data for this day
        timeSeries = convert_date(day, group['time'])
        # and then plot time series data for this day
        plt.subplot(nrows, 1, i+1)
        plt.plot(timeSeries, group['acc'], c='k')
        if imputedLabels:
            plt.fill_between(timeSeries,
                y1 = np.multiply(group['imputed'], ymax),
                y2 = np.multiply(group['imputed'], ymax * labelsPosition),
                color = labels_as_col['imputed'], alpha=1.0,
                where=group['imputed']==1)
        else:
            plt.fill(timeSeries, np.multiply(group['imputed'], ymax),
            labels_as_col['imputed'], alpha=1.0)

        # change display properties of this subplot
        ax = plt.gca()
        if len(labels)>0:
            ax.stackplot(timeSeries,
                [np.multiply(group[l], ymax * labelsPosition) for l in labels],
                colors=[labels_as_col[l] for l in labels],
                alpha=0.5, edgecolor="none")
        # add date label to left hand side of each day's activity plot
        plt.title(
            day.strftime("%A,\n%d %B"), weight='bold',
            x=-.2, y=0.5,
            horizontalalignment='left',
            verticalalignment='center',
            rotation='horizontal',
            transform=ax.transAxes,
            fontsize='medium',
            color='k'
        )
        # run gridlines for each hour bar
        ax.get_xaxis().grid(True, which='major', color='grey', alpha=0.5)
        ax.get_xaxis().grid(True, which='minor', color='grey', alpha=0.25)
        # set x and y-axes
        ax.set_xlim((datetime.combine(day,time(0, 0, 0, 0)),
            datetime.combine(day + timedelta(days=1), time(0, 0, 0, 0))))
        ax.set_xticks(pd.date_range(start=datetime.combine(day,time(0, 0, 0, 0)),
            end=datetime.combine(day + timedelta(days=1), time(0, 0, 0, 0)),
            freq='4H'))
        ax.set_xticks(pd.date_range(start=datetime.combine(day,time(0, 0, 0, 0)),
            end=datetime.combine(day + timedelta(days=1), time(0, 0, 0, 0)),
            freq='1H'), minor=True)
        ax.set_ylim((ymin, ymax))
        ax.get_yaxis().set_ticks([]) # hide y-axis lables
        # make border less harsh between subplots
        ax.spines['top'].set_color('#d3d3d3') # lightgray
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        # set background colour to lightgray
        ax.set_facecolor('#d3d3d3')

        # append to list and incrament list counter
        ax_list.append(ax)
        i += 1

    # create new subplot to display legend
    plt.subplot(nrows, 1, i+1)
    ax = plt.gca()
    ax.axis('off') # don't display axis information
    # create a 'patch' for each legend entry
    legend_patches = [mpatches.Patch(color=labels_as_col['imputed'],
                                     label='imputed', alpha=1.0),
                      mlines.Line2D([],[],color='k',label='acceleration')]
    # create lengend entry for each label
    for label in labels:
        col = labels_as_col[label]
        legend_patches.append(mpatches.Patch(color=col, label=label, alpha=0.5))
    # create overall legend
    plt.legend(handles=legend_patches, bbox_to_anchor=(0., 0., 1., 1.),
        loc='center', ncol=min(4,len(legend_patches)), mode="best",
        borderaxespad=0, framealpha=0.6, frameon=True, fancybox=True)

    # remove legend border
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax_list.append(ax)

    # format x-axis to show hours
    plt.gcf().autofmt_xdate()
    # add hour labels to top of plot
    hrLabels = ['00:00', '04:00', '08:00', '12:00', '16:00', '20:00', '24:00']
    ax_list[0].set_xticklabels(hrLabels)
    ax_list[0].tick_params(labelbottom=False, labeltop=True, labelleft=False)

    plt.savefig(plotFile, dpi=200, bbox_inches='tight')
    print('Plot file written to:', plotFile)


def str2bool(v):
    """
    Used to parse true/false values from the command line. E.g. "True" -> True
    """

    return v.lower() in ("yes", "true", "t", "1")


if __name__ == '__main__':
    main()  # Standard boilerplate to call the main() function to begin the program.



import pathlib

ROOT_DIR = pathlib.Path(__file__).parent
MODEL_DIR = ROOT_DIR / "activityModels/"
MODEL_DIR_URL = "http://gas.ndph.ox.ac.uk/aidend/accModels/"


MODELS = {
    'willetts': {
        "pth": MODEL_DIR / "willetts-jan21.tar",
        "url": MODEL_DIR_URL + "willetts-jan21.tar",
    },
    'doherty': {
        "pth": MODEL_DIR / "doherty-jan21.tar",
        "url": MODEL_DIR_URL + "doherty-jan21.tar",
    },
    'walmsley': {
        "pth": MODEL_DIR / "walmsley-jan21.tar",
        "url": MODEL_DIR_URL + "walmsley-jan21.tar",
    },
    'willetts-jan21': {
        "pth": MODEL_DIR / "willetts-jan21.tar",
        "url": MODEL_DIR_URL + "willetts-jan21.tar",
    },
    'doherty-jan21': {
        "pth": MODEL_DIR / "doherty-jan21.tar",
        "url": MODEL_DIR_URL + "doherty-jan21.tar",
    },
    'walmsley-jan21': {
        "pth": MODEL_DIR / "walmsley-jan21.tar",
        "url": MODEL_DIR_URL + "walmsley-jan21.tar",
    },
    'willetts-may20': {
        "pth": MODEL_DIR / "willetts-may20.tar",
        "url": MODEL_DIR_URL + "willetts-may20.tar",
    },
    'doherty-may20': {
        "pth": MODEL_DIR / "doherty-may20.tar",
        "url": MODEL_DIR_URL + "doherty-may20.tar",
    },
    'walmsley-nov20': {
        "pth": MODEL_DIR / "walmsley-nov20.tar",
        "url": MODEL_DIR_URL + "walmsley-nov20.tar",
    },
}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="accelerometer",
    version="4.0.0",
    author="Aiden Doherty",
    author_email="aiden.doherty@ndph.ox.ac.uk",
    description="A package to extract meaningful health information from large accelerometer datasets e.g. how much time individuals spend in sleep, sedentary behaviour, walking and moderate intensity physical activity",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/activityMonitoring/biobankAccelerometerAnalysis",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'pandas>=1.2.5',
        'scikit-learn>=0.24.2',
        'joblib',
        'statsmodels',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
    ],
    entry_points={
        "console_scripts": [
            "accProcess=accelerometer.accProcess:main",
            "accPlot=accelerometer.accPlot:main"
        ]
    },
)
